# -*- coding: utf-8 -*-
"""
    Colubrid Base Applications
    ==========================

    This file provides a list of colubrid application. Each of them inherits
    form BaseApplication and implements a full WSGI compatible web application.
    
    If you like to add your own you _have_ to inherit from BaseApplication or
    the Request object wont work properly.
"""

from __future__ import generators
from colubrid.request import Request, ResponseRequest
from colubrid.response import HttpResponse
from colubrid.utils import splitpath, fix_slash
from colubrid.const import HTTP_STATUS_CODES
from colubrid.exceptions import HttpException, PageNotFound
import inspect
import re

__all__ = ('BaseApplication', 'RegexApplication', 'ResolveRegexApplication',
           'PathApplication', 'ObjectApplication', 'WebpyApplication',
           'RoutesApplication')


class RegexCompilerClass(type):

    def __new__(cls, name, bases, d):
        result = type.__new__(cls, name, bases, d)
        if type(bases[0]) == type:
            return result
        if not hasattr(result, 'urls'):
            raise TypeError('Regex application without url definition.')
        compiled_urls = []
        for args in result.urls:
            args = list(args)
            args[0] = re.compile(args[0])
            compiled_urls.append(tuple(args))
        result.urls = compiled_urls
        return result


class RoutesMapperClass(type):

    def __new__(cls, name, bases, d):
        result = type.__new__(cls, name, bases, d)
        if type(bases[0]) == type:
            return result
        if not hasattr(result, 'mapping'):
            raise TypeError('Route application without mapping.')
        from routes import Mapper
        mapper = Mapper()
        controllers = {}
        for m in result.mapping:
            name = m[0].split('/', 1)[0]
            internal = str(id(m[1]))
            controllers[internal] = m[1]
            kwargs = {}
            if len(m) >= 3 and not m[2] is None:
                kwargs['requirements'] = m[2]
            if len(m) == 4:
                kwargs.update(m[3])
            mapper.connect(name, m[0], controller=internal, **kwargs)
        mapper.create_regs(controllers.keys())
        result._routes_mapper = mapper
        result._routes_controllers = controllers
        return result


class BaseApplication(object):

    def __init__(self, environ, start_response):
        if not hasattr(self, 'charset'):
            self.charset = 'utf-8'
        self._environ = environ
        self._start_response = start_response
        self.request = Request(environ, self.charset)
    
    def process_http_exception(self, e):
        from colubrid.exceptions import ERROR_PAGE_TEMPLATE
        from colubrid.utils import get_version
        from cgi import escape
        text = ERROR_PAGE_TEMPLATE % {
            'code':     e.code,
            'title':    escape(e.title),
            'msg':      escape(e.msg),
            'version':  get_version()
        }
        return HttpResponse(text, e.headers or [], e.code)
    
    def process_request(self):
        raise NotImplementedError
    
    def __iter__(self):
        try:
            response = self.process_request()
            if isinstance(self.request, ResponseRequest):
                response = self.request
            else:
                assert isinstance(response, HttpResponse), \
                       'process_request() must return a HttpResponse instance'
        except HttpException, e:
            response = self.process_http_exception(e)
        return response(self._start_response, self.charset)


class RegexApplication(BaseApplication):
    __metaclass__ = RegexCompilerClass

    def process_request(self):
        path_info = self.request.environ.get('PATH_INFO', '/')[1:]
        if hasattr(self, 'slash_append') and self.slash_append:
            fix_slash(self._environ, True)
        for url, module in self.urls:
            matchobj = url.search(path_info)
            if not matchobj is None:
                args = matchobj.groups()
                if module in (True, False):
                    return fix_slash(self._environ, module)
                elif not '.' in module:
                    handler = getattr(self, module)
                else:
                    p = module.split('.')
                    mname, fname = '.'.join(p[:-1]), p[-1]
                    package = __import__(mname, '', '', [''])
                    handler = getattr(package, fname)
                    args = list(args)
                    args.insert(0, self.request)
                    args = tuple(args)
                return handler(*args)
        raise PageNotFound


class ResolveRegexApplication(BaseApplication):
    __metaclass__ = RegexCompilerClass
    
    def process_request(self):
        path_info = self.request.environ.get('PATH_INFO', '/')[1:]
        if hasattr(self, 'slash_append') and self.slash_append:
            fix_slash(self._environ, True)
        for url, module in self.urls:
            matchobj = url.search(path_info)
            if not matchobj is None:
                args = matchobj.groups()
                new_args = []
                for pos, value in enumerate(args):
                    search = '$%d' % (pos + 1)
                    if search in module:
                        module = module.replace(search, value.replace('.', '_'))
                    else:
                        new_args.append(value)
                args = tuple(new_args)
                if not '.' in module:
                    if not hasattr(self, module):
                        raise PageNotFound
                    handler = getattr(self, module)
                else:
                    p = module.split('.')
                    mname, fname = '.'.join(p[:-1]), p[-1]
                    try:
                        package = __import__(mname, '', '', [''])
                        handler = getattr(package, fname)
                    except (ImportError, AttributeError):
                        raise PageNotFound
                    args = list(args)
                    args.insert(0, self.request)
                    args = tuple(args)
                if handler in (True, False):
                    return fix_slash(self._environ, handler)
                return handler(*args)
        raise PageNotFound


class WebpyApplication(BaseApplication):
    __metaclass__ = RegexCompilerClass
    
    def process_request(self):
        path_info = self.request.environ.get('PATH_INFO', '/')[1:]
        if hasattr(self, 'slash_append') and self.slash_append:
            fix_slash(self._environ, True)
        for url, cls in self.urls:
            matchobj = url.search(path_info)
            if not matchobj is None:
                cls = cls()
                cls.request = self.request
                handler = getattr(cls, self.request.environ['REQUEST_METHOD'])
                if handler in (True, False):
                    return fix_slash(self._environ, handler)
                return handler(*matchobj.groups())
        raise PageNotFound


class PathApplication(BaseApplication):

    def process_request(self):
        path_info = self.request.environ.get('PATH_INFO', '/').strip('/')
        parts = path_info.strip('/').split('/')
        if not len(parts) or not parts[0]:
            handler = 'show_index'
            args = ()
        else:
            handler = 'show_%s' % parts[0]
            args = tuple(parts[1:])
        if hasattr(self, handler):
            return getattr(self, handler)(*args)
        fix_slash(self._environ, True)
        raise PageNotFound


class ObjectApplication(BaseApplication):
    """
    a rather complex application type.
    it uses python class structures to handler the user requests.
    
    an ObjectApplication might look like this:
    
        class HelloWorld(object):
            def index(self):
                self.request.write('Hello World!')
            def name(self, name="Nobody"):
                self.request.write('Hello %s!' % name)
        
        class AdminPanel(object):
            def index(self):
                pass
            def login(self):
                pass
        
        class DispatcherApplication(ObjectApplication):
            root = HelloWorld
            root.admin = AdminPanel
            
        app = DispatcherApplication
    
    Let's say that the application listens on localhost:
    
        http://localhost/               --> HelloWorld.index()
        http://localhost/name/          --> HelloWorld.name('Nobody')
        http://localhost/name/Max       --> HelloWorld.name('Max')
        http://localhost/admin/         --> AdminPanel.index()
        http://localhost/admin/login    --> AdminPanel.login()
    """

    def process_request(self):
        if not hasattr(self, 'root'):
            raise AttributeError, 'ObjectApplication requires a root object.'
        
        path = self._environ.get('PATH_INFO', '').strip('/')
        parts = path.split('/')

        # Resolve the path
        handler = self.root
        args = []
        for part in parts:
            if part.startswith('_'):
                raise PageNotFound
            node = getattr(handler, part, None)
            if node is None:
                if part:
                    args.append(part)
            else:
                handler = node

        container = None

        # Find handler and make first container check
        if inspect.ismethod(handler):
            if handler.__name__ == 'index':
                # the handler is called index so it's the leaf of
                # itself. we don't want a slash, even if forced
                container = False
        else:
            index = getattr(handler, 'index', None)
            if not index is None:
                if not hasattr(index, 'container'):
                    container = True
                handler = index
            else:
                raise PageNotFound

        # update with hardcoded container information
        if container is None and hasattr(handler, 'container'):
            container = handler.container
        
        # Check for handler arguments and update container
        handler_args, varargs, _, defaults = inspect.getargspec(handler)
        if defaults is None:
            defaults = 0
        else:
            defaults = len(defaults)

        max_len = len(handler_args) - 1
        min_len = max_len - defaults
        cur_len = len(args)
        if varargs:
            max_len = -1

        # check if the number of arguments fits our handler
        if max_len == -1:
            if cur_len < min_len:
                raise PageNotFound
        elif min_len <= cur_len <= max_len:
            if container is None:
                container = cur_len < max_len
        else:
            raise PageNotFound

        if container is None:
            container = False
        fix_slash(self._environ, container)

        # call handler
        parent = handler.im_class()
        if hasattr(self, 'request'):
            parent.request = self.request
        return handler(parent, *args)


class RoutesApplication(BaseApplication):
    __metaclass__ = RoutesMapperClass

    def process_request(self):
        from colubrid.utils import routes_link
        path = self.request.environ.get('PATH_INFO') or '/'
        match = self._routes_mapper.match(path)
        if match is None:
            raise PageNotFound
        
        self.request.link_to = routes_link(self)
        handler = self._routes_controllers[match['controller']]
        app = handler.im_class()
        app.request = self.request
        
        if match['action'] == 'index':
            del match['action']
        del match['controller']
        return handler(app, **match)
